from typing import Any, List, Mapping, Union, cast
from dataclasses import dataclass, field

from graphql.language.ast import ListTypeNode, ListValueNode, ObjectValueNode


from graphql import (
    GraphQLSchema,
    validate,
    parse,
    get_operation_ast,
    visit,
    Visitor,
    TypeInfo,
    TypeInfoVisitor,
    GraphQLNonNull,
    is_scalar_type,
    GraphQLList,
    OperationDefinitionNode,
    NonNullTypeNode,
    TypeNode,
    GraphQLEnumType,
    is_enum_type,
)


@dataclass
class ParsedField:
    name: str
    type: str
    nullable: bool
    default_value: Any = None


@dataclass
class ParsedObject:
    name: str
    fields: List[ParsedField] = field(default_factory=list)
    parents: List[str] = field(default_factory=list)
    children: List["ParsedObject"] = field(default_factory=list)


@dataclass
class ParsedEnum:
    name: str
    values: Mapping[str, Any]


@dataclass
class ParsedVariableDefinition:
    name: str
    type: str
    nullable: bool
    default_value: Any = None


@dataclass
class ParsedOperation:
    name: str
    type: str
    variables: List[ParsedVariableDefinition] = field(default_factory=list)
    children: List[ParsedObject] = field(default_factory=list)


NodeT = Union[ParsedOperation, ParsedObject]


@dataclass
class ParsedQuery:
    query: str
    objects: List[NodeT] = field(default_factory=list)
    enums: List[ParsedEnum] = field(default_factory=list)


class FieldToTypeMatcherVisitor(Visitor):

    def __init__(self, schema: GraphQLSchema, type_info: TypeInfo, query: str):
        self.schema = schema
        self.type_info = type_info
        self.query = query
        self.parsed = ParsedQuery(query=self.query)
        self.dfs_path: List[ParsedObject] = []

    def push(self, obj: NodeT):
        self.dfs_path.append(obj)

    def pull(self) -> NodeT:
        return self.dfs_path.pop()

    @property
    def current(self) -> ParsedObject:
        return self.dfs_path[-1]

    def extract_default_value(self, value_node):
        """Extracts the default value from a value node."""
        if hasattr(value_node, 'value'):  # Directly accessible value (scalar types)
            return value_node.value
        elif isinstance(value_node, ListValueNode):  # List values
            return [self.extract_default_value(item) for item in value_node.values]
        elif isinstance(value_node, ObjectValueNode):  # Object values
            return {field.name.value: self.extract_default_value(field.value) for field in value_node.fields}
        # Add other types as needed
        return None

    # Document
    def enter_operation_definition(self, node: OperationDefinitionNode, *_args):
        name, operation = node.name, node.operation

        variables = []
        for var in node.variable_definitions:
            ptype, nullable, _ = self.__variable_type_to_python(var.type)
            default_value = self.extract_default_value(var.default_value) if var.default_value else None
            variables.append(
                ParsedVariableDefinition(
                    name=var.variable.name.value,
                    type=ptype,
                    nullable=nullable,
                    default_value=default_value,
                )
            )

        parsed_op = ParsedOperation(
            name=name.value,
            type=str(operation.value),
            variables=variables,
            children=[ParsedObject(name=f"{name.value}Data")],
        )

        self.parsed.objects.append(parsed_op)  # pylint:disable=no-member
        self.push(parsed_op)
        self.push(parsed_op.children[0])  # pylint:disable=unsubscriptable-object

        return node


    # def enter_selection_set(self, node, *_):
    #     return node

    def leave_selection_set(self, node, *_):
        self.pull()
        return node

    # Fragments

    def enter_fragment_definition(self, node, *_):
        # Same as operation definition
        obj = ParsedObject(name=node.name.value)
        self.parsed.objects.append(obj)  # pylint:disable=no-member
        self.push(obj)
        return node

    def enter_fragment_spread(self, node, *_):
        self.current.parents.append(node.name.value)
        return node

    # def enter_inline_fragment(self, node, *_):
    #     return node
    #
    # def leave_inline_fragment(self, node, *_):
    #     return node

    # Field

    def enter_field(self, node, *_):
        name = node.alias.value if node.alias else node.name.value
        graphql_type = self.type_info.get_type()
        python_type, nullable, underlying_graphql_type = self.__scalar_type_to_python(
            graphql_type
        )

        parsed_field = ParsedField(
            name=name,
            type=python_type,
            nullable=nullable,
        )

        self.current.fields.append(parsed_field)  # TODO: nullables should go to the end

        if not is_scalar_type(underlying_graphql_type):
            if is_enum_type(underlying_graphql_type):
                enum_type = cast(
                    GraphQLEnumType, self.schema.type_map[underlying_graphql_type.name]
                )
                name = enum_type.name
                if not any(
                    e.name == name for e in self.parsed.enums
                ):  # pylint:disable=not-an-iterable
                    parsed_enum = ParsedEnum(
                        name=enum_type.name,
                        values={
                            name: value.value or name
                            for name, value in enum_type.values.items()
                        },
                    )
                    self.parsed.enums.append(parsed_enum)  # pylint:disable=no-member
            else:
                obj = ParsedObject(name=str(underlying_graphql_type))

                self.current.children.append(obj)
                self.push(obj)

        return node

    @staticmethod
    def __scalar_type_to_python(scalar):
        nullable = True
        if isinstance(scalar, GraphQLNonNull):
            nullable = False
            scalar = scalar.of_type

        mapping = {
            "ID": "str",
            "String": "str",
            "Int": "int",
            "bigint": "int",
            "Float": "float",
            "Boolean": "bool",
            "DateTime": "DateTime",
            "timestamptz": "DateTime",
        }

        if isinstance(scalar, GraphQLList):
            scalar = scalar.of_type
            if isinstance(scalar, GraphQLNonNull):
                scalar = scalar.of_type
                nullable = False
            mapped_type = mapping.get(str(scalar), str(scalar))
            mapping = f"List[{mapped_type}]"
        else:
            mapping = mapping.get(str(scalar), str(scalar))

        return mapping, nullable, scalar


    @staticmethod
    def __variable_type_to_python(var_type: TypeNode):
        nullable = True
        original_var_type = var_type  # Keep the original type for mapping later
        
        # Recursively navigate through NonNullTypeNode and ListTypeNode to find the NamedTypeNode
        while isinstance(var_type, (NonNullTypeNode, ListTypeNode)):
            if isinstance(var_type, NonNullTypeNode):
                nullable = False
                var_type = var_type.type  # Go one level deeper
            elif isinstance(var_type, ListTypeNode):
                var_type = var_type.type  # Go one level deeper

        # Now, var_type should be a NamedTypeNode, and we can safely access its name
        mapping = {
            "ID": "str",
            "String": "str",
            "Int": "int",
            "Float": "float",
            "Boolean": "bool",
            "DateTime": "DateTime",
        }

        # Handle lists by checking the original_var_type
        if isinstance(original_var_type, ListTypeNode):
            type_str = f"List[{mapping.get(var_type.name.value, var_type.name.value)}]"
        else:
            type_str = mapping.get(var_type.name.value, var_type.name.value)

        return type_str, nullable, var_type


class AnonymousQueryError(Exception):
    def __init__(self):
        super().__init__("All queries must be named")


class InvalidQueryError(Exception):
    def __init__(self, errors):
        self.errors = errors
        message = "\n".join(str(err) for err in errors)
        super().__init__(message)


class QueryParser:
    def __init__(self, schema: GraphQLSchema):
        self.schema = schema
        self.__jinja2_env = None

    def parse(self, query: str, should_validate: bool = True) -> ParsedQuery:
        document_ast = parse(query)
        operation = get_operation_ast(document_ast)

        if not operation.name:
            raise AnonymousQueryError()

        if should_validate:
            errors = validate(self.schema, document_ast)
            if errors:
                raise InvalidQueryError(errors)

        type_info = TypeInfo(self.schema)
        visitor = FieldToTypeMatcherVisitor(self.schema, type_info, query)
        visit(document_ast, TypeInfoVisitor(type_info, visitor))
        result = visitor.parsed
        return result



import os
import json
import requests
from graphql import get_introspection_query, build_client_schema


def load_introspection_from_server(url):
    query = get_introspection_query()
    request = requests.post(url, json={'query': query})
    if request.status_code == 200:
        return request.json()['data']

    raise Exception(f'Query failed to run by returning code of {request.status_code}. {query}')


def load_introspection_from_file(filename):
    with open(filename, 'r') as fin:
        return json.load(fin)


def load_schema(uri):
    introspection = load_introspection_from_file(uri) if os.path.isfile(uri) else load_introspection_from_server(uri)
    return build_client_schema(introspection)



#!/usr/bin/env python
import textwrap
import click
import glob
import subprocess
import time
import os
import pluralizer
import re

from os.path import join as join_paths, isfile, dirname, basename, splitext

from collections import defaultdict
from graphql import GraphQLSchema
from watchdog.observers import Observer
from watchdog.events import (
    FileSystemEventHandler,
    EVENT_TYPE_CREATED,
    EVENT_TYPE_MODIFIED,
)
from gql.utils_codegen import CodeChunk

from gql.config import Config
from gql.query_parser import QueryParser, AnonymousQueryError, InvalidQueryError
from gql.renderer_dataclasses import DataclassesRenderer
from gql.utils_schema import load_schema

DEFAULT_CONFIG_FNAME = ".gql.json"
SCHEMA_PROMPT = click.style("Where is your schema?: ", fg="bright_white") + click.style(
    "(path or url) ", fg="bright_black", dim=False
)

ROOT_PROMPT = click.style(
    "Whats the root of your project: ", fg="bright_white"
) + click.style("(path or url) ", fg="bright_black", dim=False)


def safe_remove(fname):
    try:
        os.remove(fname)
    except:
        pass


@click.group()
def cli():
    pass


@cli.command()
@click.option("--schema", prompt=SCHEMA_PROMPT, default="http://localhost:4000")
@click.option("--endpoint", prompt=SCHEMA_PROMPT, default="same as schema")
@click.option("--root", prompt=ROOT_PROMPT, default="./src")
@click.option(
    "-c",
    "--config",
    "config_filename",
    default=DEFAULT_CONFIG_FNAME,
    type=click.Path(exists=False),
)
def init(schema, endpoint, root, config_filename):
    if isfile(config_filename):
        click.confirm(
            f"{config_filename} already exists. Are you sure you want to continue?",
            abort=True,
        )

    if endpoint == "same as schema":
        endpoint = schema

    config = Config(
        schema=schema,
        endpoint=endpoint,
        documents=join_paths("queries", "**/*.graphql"),
    )

    config.save(config_filename)

    click.echo(
        f"Config file generated at {click.style(config_filename, fg='bright_white')}\n\n"
    )


def process_files_with_same_domain(
    filenames: list, parser: QueryParser, renderer: DataclassesRenderer
):
    def to_snake_case(camel_case_string):
        # Use regular expression to identify capital letters and add underscores before them
        snake_case_string = re.sub(r"([a-z0-9])([A-Z])", r"\1_\2", camel_case_string)
        # Convert the entire string to lowercase
        return snake_case_string.lower()

    grouped_files = {
        "app": {},
        "packages": {},
    }
    for filename in filenames:
        parts = filename.split("/")

        del parts[0]

        name_of_the_file = parts[-1]
        app_or_package = parts[0]

        name_of_the_model = ""
        if app_or_package == "app":
            name_of_the_model = parts[2]
        elif app_or_package == "packages":
            name_of_the_model = parts[1]

        if name_of_the_model not in grouped_files[app_or_package]:
            grouped_files[app_or_package][name_of_the_model] = []

        grouped_files[app_or_package][name_of_the_model].append(
            {
                "full_path": filename,
                "name_of_the_file": name_of_the_file,
                "name_of_the_function": to_snake_case(
                    name_of_the_file.split(".graphql")[0]
                ),
                "app_or_package": app_or_package,
                "name_of_the_model": name_of_the_model,
            }
        )

    packages_data = {}
    for key, items in grouped_files["packages"].items():
        for item in items:
            needle = item["full_path"].split("/")[4]

            if needle not in packages_data:
                packages_data[needle] = []

            packages_data[needle].append(item)
        grouped_files["packages"][key] = packages_data

    gathered_data = {}

    for key, models in grouped_files["app"].items():
        for item in models:
            filename, function = process_files_in_directory(
                full_path=item["full_path"],
                name_of_the_file=item["name_of_the_file"],
                app_or_package=item["app_or_package"],
                name_of_the_function=item["name_of_the_function"],
                name_of_the_model=item["name_of_the_model"],
                parser=parser,
                renderer=renderer,
            )

            if key not in gathered_data:
                gathered_data[key] = {
                    "domain_type": "app",
                    "path": os.path.normpath(os.path.join(item["full_path"], "../..")),
                    "functions": [],
                }

            gathered_data[key]["functions"].append(
                {
                    "filename": filename,
                    "function": function,
                }
            )

    for parent_name, boks in grouped_files["packages"].items():
        for key, items in boks.items():
            for item in items:
                filename, function = process_files_in_directory(
                    full_path=item["full_path"],
                    name_of_the_file=item["name_of_the_file"],
                    app_or_package=item["app_or_package"],
                    name_of_the_function=item["name_of_the_function"],
                    name_of_the_model=item["name_of_the_model"],
                    parser=parser,
                    renderer=renderer,
                )

                if parent_name + key not in gathered_data:
                    gathered_data[parent_name + key] = {
                        "domain_type": "packages",
                        "actual_name": key,
                        "path": os.path.normpath(
                            os.path.join(item["full_path"], "../..")
                        ),
                        "functions": [],
                    }

                gathered_data[parent_name + key]["functions"].append(
                    {
                        "filename": filename,
                        "function": function,
                    }
                )

    for name_of_the_model, item in gathered_data.items():
        create_init_file(
            name_of_the_model=name_of_the_model,
            directory=item["path"],
            functions=item["functions"],
            domain_type=item["domain_type"],
            actual_name=item.get("actual_name", None),
        )


def create_init_file(
    name_of_the_model: str,
    directory: str,
    functions: list,
    domain_type: str,
    actual_name: str = None,
):
    init_file_path = os.path.join(directory, "__init__.py")

    # List all files in the directory
    files_in_directory = os.listdir(directory)

    # Create or overwrite the __init__.py file
    with open(init_file_path, "w") as init_file:
        init_file.write("# pyright: reportUnusedImport=false\n")

        # Include additional functions specified in the functions list
        for item in functions:
            init_file.write(
                f"from .executor.{item['filename']} import {item['function']}\n"
            )

        # Automatically include files at the same level
        for file_name in files_in_directory:
            if file_name.endswith(".py") and file_name != "__init__.py":
                class_name = os.path.splitext(file_name)[0]
                init_file.write(f"from .{class_name} import {class_name}\n")


def process_files_in_directory(
    full_path: str,
    name_of_the_file: str,
    app_or_package: str,
    name_of_the_model: str,
    name_of_the_function: str,
    parser: QueryParser,
    renderer: DataclassesRenderer,
):
    bare_file_name = "_".join(name_of_the_file.split(".")[:-1])
    verb = name_of_the_function.split("_")[0]

    domain_name = re.sub(r"(?<!^)(?=[A-Z])", "_", name_of_the_model).lower()

    start_path = os.path.normpath(os.path.join(full_path, "../.."))
    output_file = f"{start_path}/executor/{bare_file_name}.py"

    buffer = CodeChunk()
    buffer.write(renderer.render_shared_code())

    click.echo(f"Parsing {full_path} ... ", nl=False)

    with open(full_path, "r") as fin:
        query = fin.read()
        try:
            parsed = parser.parse(query)
            rendered = renderer.render(parsed, full_path, verb)

            with buffer.write_block("def {}():".format(name_of_the_function)):
                buffer.write("return {}, '{}'".format(verb, parsed.objects[0].name))

            buffer.write(f"class {verb}:")
            buffer.write("    " + rendered.replace("\n", "\n    "))
            click.secho("Success!", fg="bright_white")
        except AnonymousQueryError:
            click.secho("Failed!", fg="bright_red")
            click.secho("\tQuery is missing a name", fg="bright_black")
        except InvalidQueryError as invalid_err:
            click.secho("Failed!", fg="bright_red")
            click.secho(f"\t{invalid_err}", fg="bright_black")

    if len(buffer.lines) > 2:
        os.makedirs(dirname(output_file), exist_ok=True)
        with open(output_file, "w") as outfile:
            for chunk in buffer.lines:
                outfile.write(chunk)
                outfile.write("\n")

        # Format the output file using Black
        format_with_black(output_file)

    return bare_file_name, name_of_the_function


@cli.command()
@click.option(
    "-c",
    "--config",
    "config_filename",
    default=DEFAULT_CONFIG_FNAME,
    type=click.Path(exists=True),
)
def run(config_filename):
    if not isfile(config_filename):
        click.echo(f"Could not find configuration file {config_filename}")

    config = Config.load(config_filename)
    schema = load_schema(config.schema)

    filenames = glob.glob(config.documents, recursive=True)

    query_parser = QueryParser(schema)
    query_renderer = DataclassesRenderer(schema, config)

    process_files_with_same_domain(filenames, query_parser, query_renderer)


def format_with_black(filename):
    """Format a Python file using Black."""
    try:
        subprocess.run(["black", filename], check=True)
        click.echo(f"Formatted {filename} with Black.")
    except subprocess.CalledProcessError as e:
        click.echo(f"Error formatting {filename} with Black: {e}")


@cli.command()
@click.option(
    "-c",
    "--config",
    "config_filename",
    default=DEFAULT_CONFIG_FNAME,
    type=click.Path(exists=True),
)
def watch(config_filename):
    class Handler(FileSystemEventHandler):
        def __init__(self, config: Config, schema: GraphQLSchema):
            self.parser = QueryParser(schema)
            self.renderer = DataclassesRenderer(schema, config)

        def on_any_event(self, event):
            if event.is_directory:
                return

            if event.event_type in {EVENT_TYPE_CREATED, EVENT_TYPE_MODIFIED}:
                filenames = [
                    os.path.abspath(fn)
                    for fn in glob.iglob(config.documents, recursive=True)
                ]
                if event.src_path not in filenames:
                    return

                process_files_with_same_domain(filenames, self.parser, self.renderer)

    if not isfile(config_filename):
        click.echo(f"Could not find configuration file {config_filename}")

    config = Config.load(config_filename)
    schema = load_schema(config.schema)

    click.secho(f"Watching {config.documents}", fg="cyan")
    click.secho("Ready for changes...", fg="cyan")

    observer = Observer()
    observer.schedule(Handler(config, schema), os.path.abspath("./"), recursive=True)
    observer.start()
    try:
        while True:
            time.sleep(5)
    except:
        observer.stop()
        print("Error")

    observer.join()


if __name__ == "__main__":
    cli()



from typing import Callable, Mapping, Union

import aiohttp


class AsyncIOClient:
    def __init__(self, endpoint, headers=None):
        self.endpoint = endpoint

        headers = headers or {}
        self.__headers = {
            **headers,
            'Content-Type': 'application/json',
            'Accept': 'application/json',
            'Accept-Encoding': 'gzip',
        }
        self.__session = None

    @property
    def session(self):
        if not self.__session or self.__session.closed:
            self.__session = aiohttp.ClientSession(headers=self.__headers)

        return self.__session

    async def call(self, query,
                   variables=None,
                   return_json=False,
                   on_before_callback: Callable[[Mapping[str, str], Mapping[str, str]], None] = None) -> Union[dict, str]:

        headers = self.__headers.copy()

        payload = {
            'query': query
        }
        if variables:
            payload['variables'] = variables

        if on_before_callback:
            on_before_callback(payload, headers)

        async with self.session.post(self.endpoint, json=payload, headers=headers, raise_for_status=True) as resp:
            if return_json:
                return await resp.json()

            return await resp.text()



from graphql import GraphQLSchema

from gql.config import Config
from gql.utils_codegen import CodeChunk
from gql.query_parser import (
    ParsedQuery,
    ParsedField,
    ParsedObject,
    ParsedEnum,
    ParsedOperation,
    ParsedVariableDefinition,
)

from typing import Optional, Dict, Any

class DataclassesRenderer:

    def __init__(self, schema: GraphQLSchema, config: Config):
        self.schema = schema
        self.config = config

    def render(self, parsed_query: ParsedQuery):
        # We sort fragment nodes to be first and operations to be last because of dependecies
        buffer = CodeChunk()
        buffer.write("# AUTOGENERATED file. Do not Change!")
        buffer.write("import re")
        buffer.write("from functools import partial")
        buffer.write("from typing import Any, Callable, Mapping, List")
        buffer.write("from enum import Enum")
        buffer.write("from dataclasses import dataclass, field")
        buffer.write("from dataclasses_json import dataclass_json")
        buffer.write("from gql.clients import Client, AsyncIOClient")
        buffer.write("from gql.renderer_dataclasses import DataclassesRenderer")
        buffer.write("from masonite.facades import Auth")
        buffer.write("")

        if self.config.custom_header:
            buffer.write_lines(self.config.custom_header.split("\n"))

        buffer.write("")

        self.__render_datetime_field(buffer)

        # Enums
        if parsed_query.enums:
            self.__render_enum_field(buffer)
            for enum in parsed_query.enums:
                self.__render_enum(buffer, enum)

        sorted_objects = sorted(
            parsed_query.objects,
            key=lambda obj: 1 if isinstance(obj, ParsedOperation) else 0,
        )
        for obj in sorted_objects:
            if isinstance(obj, ParsedObject):
                self.__render_object(parsed_query, buffer, obj)
            elif isinstance(obj, ParsedOperation):
                self.__render_operation(parsed_query, buffer, obj)

        return str(buffer)

    @staticmethod
    def __render_enum_field(buffer: CodeChunk):
        with buffer.write_block("def enum_field(enum_type):"):
            with buffer.write_block("def encode_enum(value):"):
                buffer.write("return value.value")
            buffer.write("")
            with buffer.write_block("def decode_enum(t, value):"):
                buffer.write("return t(value)")

            buffer.write("")
            buffer.write(
                "return field(metadata={'dataclasses_json': {'encoder': encode_enum, 'decoder': partial(decode_enum, enum_type)}})"
            )
            buffer.write("")

    @staticmethod
    def __render_datetime_field(buffer: CodeChunk):
        buffer.write("")
        buffer.write("from datetime import datetime")
        buffer.write("from marshmallow import fields as marshmallow_fields")
        buffer.write(
            "DATETIME_FIELD = field(metadata={'dataclasses_json': {'encoder': datetime.isoformat, 'decoder': datetime.fromisoformat, 'mm_field': marshmallow_fields.DateTime(format='iso')}})"
        )
        buffer.write("")

    def __render_object(
        self, parsed_query: ParsedQuery, buffer: CodeChunk, obj: ParsedObject
    ):
        class_parents = "" if not obj.parents else f'({", ".join(obj.parents)})'
        
        buffer.write("@dataclass_json")
        buffer.write("@dataclass")
        with buffer.write_block(f"class {obj.name}{class_parents}:"):
            # render child objects
            for child_object in obj.children:
                self.__render_object(parsed_query, buffer, child_object)

            # render fields
            sorted_fields = sorted(
                obj.fields, key=lambda f: (f.type != "DateTime", f.nullable)
            )
            for field in sorted_fields:
                self.__render_field(parsed_query, buffer, field)

            # pass if not children or fields
            if not (obj.children or obj.fields):
                buffer.write("pass")

        buffer.write("")
        
    @staticmethod
    def simplify(response: Dict[str, Any], operation_name: str) -> Dict[str, Any]:
        if 'data' not in response or ('errors' in response and response['errors']):
            return response  # Return as-is in case of errors or missing data
        
        # if operation_name is store, make it insert
        operation_name = operation_name.replace('store', 'insert')
        operation_name = operation_name.replace('destroy', 'delete')
        
        data = response['data']
        base_key = operation_name
        item_key = f"{base_key}"

        if f"{item_key}_by_pk" in data:
            data = {**data, **data[f"{item_key}_by_pk"]}
            data.pop(f"{item_key}_by_pk")
            
            return data
        
        return data.get(item_key, [])

    @staticmethod
    def relayify(response: Dict[str, Any], operation_name: str) -> Dict[str, Any]:
        """
        Transforms a standard GraphQL response into a Relay-style response.
        Dynamically handles based on operation_name extracted from class name or passed explicitly.
        """
        if 'data' not in response or ('errors' in response and response['errors']):
            return response  # Return as-is in case of errors or missing data
        
        # if operation_name is store, make it insert
        operation_name = operation_name.replace('store', 'insert')
        operation_name = operation_name.replace('destroy', 'delete')
        
        data = response['data']
        base_key = operation_name  # Adjust this if you need to extract from class name or metadata
        item_key = f"{base_key}"  # Adjust according to your data structure, e.g., for `amazon_product` queries
        aggregate_key = f"{base_key}_aggregate"
        
        total_count = data.get(aggregate_key, {}).get('aggregate', {}).get('count', None)
        
        relay_response = {}

        if total_count is not None:
            relay_response["totalCount"] = total_count
            relay_response["pageInfo"] = {
                "hasNextPage": False,  # Placeholder, adjust based on your pagination logic
                "endCursor": None,  # Placeholder for cursor logic
            }
            if total_count > 0:
                edges = [
                    {
                        "cursor": str(index),  # Placeholder for cursor logic
                        "node": item,
                    }
                    for index, item in enumerate(data.get(item_key, []), start=1)
                ]
                relay_response["edges"] = edges
                return relay_response
        
        if f"{item_key}_by_pk" in data:
            data = {**data, **data[f"{item_key}_by_pk"]}
            data.pop(f"{item_key}_by_pk")
        
        # if response contains more than one item, return as-is
        if len(data) > 1:
            return data
        
        return data.get(item_key, [])
    
    def __render_operation(
        self, parsed_query: ParsedQuery, buffer: CodeChunk, parsed_op: ParsedOperation
    ):
        buffer.write("@dataclass_json")
        buffer.write("@dataclass")
        with buffer.write_block(f"class {parsed_op.name}:"):
            buffer.write('__QUERY__ = """')
            buffer.write(parsed_query.query)
            buffer.write('"""')
            buffer.write("")

            # Render children
            for child_object in parsed_op.children:
                self.__render_object(parsed_query, buffer, child_object)

            # operation fields
            buffer.write("")
            buffer.write(f"data: {parsed_op.name}Data = None")
            buffer.write("errors: Any = None")
            buffer.write("")

            # Execution functions
            vars_args_list = [f"{var.name}=None" for var in parsed_op.variables]
            vars_args_str = ", ".join(vars_args_list)
            if vars_args_str:
                vars_args_str += ", "

            variables_dict_lines = []
            for x in parsed_op.variables:
                defval = x.default_value
                if x.type == "str" and defval is None:
                    defval = '""'

                variables_dict_lines.append(
                    f"'{x.name}': {x.name} if {x.name} is not None else {defval}"
                )
                
            buffer.write("@classmethod")
            with buffer.write_block(
                f"def execute(cls, {vars_args_str}on_before_callback: Callable[[Mapping[str, str], Mapping[str, str]], None] = None):"
            ):
                buffer.write(f"client = Client('{self.config.endpoint}')")
                variables_dict_str = ", ".join(variables_dict_lines)
                buffer.write(f"variables = {{{variables_dict_str}}}")
                buffer.write(
                    "response_text = client.call(cls.__QUERY__, variables=variables, on_before_callback=on_before_callback)"
                )
                buffer.write("data = cls.from_json(response_text).to_dict()")

                buffer.write("operation_name = cls.__name__.replace('find', '').replace('get', '')")
                buffer.write("singular_str = operation_name[:-1] if operation_name.endswith('s') else operation_name")
                buffer.write("snake_case_str = re.sub(r'(?<!^)(?=[A-Z])', '_', singular_str).lower()")
                
                buffer.write("authenticated_user = Auth.user()")
                buffer.write("return DataclassesRenderer.relayify(data, snake_case_str) if authenticated_user else DataclassesRenderer.simplify(data, snake_case_str)")

            buffer.write("")
            
            buffer.write("@classmethod")
            with buffer.write_block(
                f"async def execute_async(cls, {vars_args_str}on_before_callback: Callable[[Mapping[str, str], Mapping[str, str]], None] = None):"
            ):
                buffer.write(f"client = AsyncIOClient('{self.config.endpoint}')")
                buffer.write(f"variables = {{{variables_dict_str}}}")
                buffer.write(
                    "response_text = await client.call(cls.__QUERY__, variables=variables, on_before_callback=on_before_callback)"
                )
                buffer.write("data = cls.from_json(response_text).to_dict()")

                buffer.write("operation_name = cls.__name__.replace('find', '').replace('get', '')")
                buffer.write("singular_str = operation_name[:-1] if operation_name.endswith('s') else operation_name")
                buffer.write("snake_case_str = re.sub(r'(?<!^)(?=[A-Z])', '_', singular_str).lower()")
                
                buffer.write("authenticated_user = Auth.user()")
                buffer.write("return DataclassesRenderer.relayify(data, snake_case_str) if authenticated_user else DataclassesRenderer.simplify(data, snake_case_str)")

            buffer.write("")
            buffer.write("")

    @staticmethod
    def __render_variable_definition(var: ParsedVariableDefinition):
        if not var.nullable:
            return f"{var.name}: {var.type}"

        return f'{var.name}: {var.type} = {var.default_value or "None"}'

    @staticmethod
    def __render_field(
        parsed_query: ParsedQuery, buffer: CodeChunk, field: ParsedField
    ):
        enum_names = [e.name for e in parsed_query.enums]
        is_enum = field.type in enum_names
        suffix = ""
        field_type = field.type

        if is_enum:
            suffix = f"= enum_field({field.type})"

        if field.nullable:
            suffix = f"= {field.default_value}"

        if field.type == "DateTime":
            suffix = "= DATETIME_FIELD"
            field_type = "datetime"
        
        if field_type in ["List[user_amazon_association]", "List[user_ebay_association]"]:
            suffix = f"= {field.default_value}"

        buffer.write(f"{field.name}: {field_type} {suffix}")

    @staticmethod
    def __render_enum(buffer: CodeChunk, enum: ParsedEnum):
        with buffer.write_block(f"class {enum.name}(Enum):"):
            for value_name, value in enum.values.items():
                if isinstance(value, str):
                    value = f"'{value}'"

                buffer.write(f"{value_name} = {value}")

        buffer.write("")



from typing import Callable, Mapping, Union

import requests


class Client:
    def __init__(self, endpoint, headers=None):
        self.endpoint = endpoint

        headers = headers or {}
        self.__headers = {
            **headers,
            'Content-Type': 'application/json',
            'Accept': 'application/json',
            'Accept-Encoding': 'gzip',
        }

    def call(self, query,
             variables=None,
             return_json=False,
             on_before_callback: Callable[[Mapping[str, str], Mapping[str, str]], None] = None) -> Union[dict, str]:

        headers = self.__headers.copy()

        payload = {
            'query': query
        }
        if variables:
            payload['variables'] = variables

        if on_before_callback:
            on_before_callback(payload, headers)

        response = requests.post(self.endpoint, json=payload, headers=headers)
        response.raise_for_status()
        return response.json() if return_json else response.text



from typing import Type, TypeVar
from dataclasses import dataclass
from dataclasses_json import dataclass_json


ConfigT = TypeVar('ConfigT', bound='ConfigT')

@dataclass_json
@dataclass(frozen=True)
class Config:
    schema: str
    endpoint: str
    documents: str
    custom_header: str = ''

    @classmethod
    def load(cls: Type[ConfigT], filename: str) -> ConfigT:
        with open(filename, 'r') as fin:
            json_str = fin.read()
            return cls.from_json(json_str)  # pylint:disable=no-member

    def save(self, filename, pretty=True):
        with open(filename, 'w') as outfile:
            json_str = self.to_json(indent=2) if pretty else self.to_json()  # pylint:disable=no-member
            outfile.write(json_str)



from setuptools import setup

setup(
    name="cfk_gql",
    version="5.0",
    packages=["gql", "gql.clients"],
    package_data={"": ["*"]},
    package_dir={"": "."},
    install_requires=[
        "click",
        "GraphQL-core-next",
        "watchdog",
        "requests",
        "aiohttp",
        "dataclasses_json",
    ],
    entry_points={
        "console_scripts": [
            "gql = gql.cli:cli",
        ],
    },
    # Additional metadata
    author="Lio",
    author_email="lio@example.com",
)
